import modelindex


def test_model_markdown():
    mi = modelindex.load("tests/test-mi/11_markdown")

    assert len(mi.models) == 6
    assert mi.models[0].name == "RexNet"
    assert mi.models[0].metadata.data["some field"] == 10

    assert mi.models[1].name == "RexNet1"
    assert mi.models[1].metadata.data["some field"] == 11
    assert mi.models[2].name == "RexNet2"
    assert mi.models[2].metadata.data["some field"] == 22
    assert mi.models[3].name == "RexNet3"
    assert mi.models[3].results[0].metrics == {"mAP": "19%"}
    assert mi.models[4].name == "Inception v3 - 90 epochs"
    assert mi.models[5].name == "Inception v3 - 120 epochs"



import os
import copy
from typing import Dict, Union, List

from ordered_set import OrderedSet

from modelindex.models.Metadata import Metadata
from modelindex.models.BaseModelIndex import BaseModelIndex
from modelindex.models.Result import Result
from modelindex.models.ResultList import ResultList
from modelindex.utils import lowercase_keys, full_filepath, load_any_file, expand_wildcard_path, merge_lists_data


class Model(BaseModelIndex):
    """Model represents the ML model.
    """
    COMMON_FIELDS = [
        "Name",
        "Metadata",
        "Results",
        "Paper",
        "Code",
        "Weights",
        "Config",
        "README",
        "In Collection",
        "Image",
    ]

    def __init__(self,
                 name: str = None,
                 metadata: Union[Dict, Metadata, str] = None,
                 results: Union[List, ResultList, Result, Dict, str] = None,
                 paper: Union[str, Dict] = None,
                 code: str = None,
                 weights: str = None,
                 config: str = None,
                 readme: str = None,
                 in_collection: Union[str, List[str]] = None,
                 image: str = None,
                 _filepath: str = None,
                 _path_to_readme: str=None,
                 **kwargs,
                 ):
        """
        Args:
            name (str): Name of the model
            metadata (Metadata, dict, str): Metadata object, metadata dict or a filepath to the metadata file
            results (ResultList, Result, list, dict, str): ResultList, a single Results, a list of result dicts, a single
                                                     result dict, or a filepath to the result file
            paper (str, dict): URL to the paper, or a structured dict with paper metadata (title, url)
            code (str): URL to the code snippet
            weights (str): URL to the pretrained weights
            config (str): URL to the config file
            readme (str): path to the README file for the model
            in_collection (str, List): name of the collection to which the model belongs to
            image (str): path or URL to an image for the model
            _filepath: The file path to where the data was loaded from
            _path_to_readme: Path to the markdown readme file if data is coming from there
            **kwargs: Any other custom fields
        """

        check_errors = OrderedSet()

        if metadata is not None and isinstance(metadata, str):
            # link to a metadata file
            try:
                metadata = Metadata.from_file(metadata, _filepath)
            except (IOError, ValueError) as e:
                check_errors.add(str(e))
        elif metadata is not None and not isinstance(metadata, Metadata):
            metadata = Metadata.from_dict(metadata, _filepath)

        if results is not None and isinstance(results, str):
            # link to 1+ result files
            results_list = []
            for results_file in expand_wildcard_path(results, _filepath):
                try:
                    results_list.append(ResultList.from_file(results_file, _filepath))
                except (IOError, ValueError) as e:
                    check_errors.add(str(e))
            results = merge_lists_data(results_list)
        elif results is not None and not isinstance(results, ResultList):
            results = ResultList(results, _filepath)

        d = {
            "Name": name,
            "Metadata": metadata,
            "Results": results,
            "Paper": paper,
            "Code": code,
            "Weights": weights,
            "Config": config,
            "README": readme,
            "In Collection": in_collection,
            "Image": image,
            **kwargs,
        }

        # Only non-empty items
        data = {k: v for k, v in d.items() if v is not None}

        self._path_to_readme = _path_to_readme
        self._full_model = self

        super().__init__(
            data=data,
            filepath=_filepath,
            check_errors=check_errors,
        )

    def _check(self, silent=True):
        if self.name is None or self.name == "":
            self.check_errors.add("Field 'Name' cannot be empty")

        if self._readme_is_filepath() and not self._path_to_readme:
            # check if the README exists
            fullpath = full_filepath(self.readme, self.filepath)
            if not os.path.isfile(fullpath):
                self.check_errors.add(f"Path to README file {self.readme} is not a valid file.")

        if self.image and not self.image.startswith("http"):
            fullpath = full_filepath(self.image, self.filepath)
            if not os.path.isfile(fullpath):
                self.check_errors.add(f"Path to Image file {self.image} is not a valid file.")

    @classmethod
    def from_dict(cls, d: Dict, _filepath: str = None, _path_to_readme: str = None):
        """Create a Model from a dictionary.

        Args:
            d (dict): dictionary containing models data
            _filepath (str): The file path to where the data was loaded from
            _path_to_readme (str): Path to the README file if metadata was extracted from a README
        """
        lc_keys = lowercase_keys(d)

        copy_fields = [
            "name",
            "paper",
            "code",
            "weights",
            "config",
            "readme",
            "metadata",
            "results",
            "in_collection",
            "image",
        ]

        dd = d.copy()
        for field_name in copy_fields:
            key = field_name.lower()
            if key in lc_keys:
                dd[field_name] = dd.pop(lc_keys[key])

            # try with _ instead of space in the field name
            if " " in field_name:
                key = field_name.lower().replace(" ", "_")
                if key in lc_keys:
                    dd[field_name] = dd.pop(lc_keys[key])

        if _path_to_readme:
            dd["readme"] = _path_to_readme

        return cls(
            _filepath=_filepath,
            _path_to_readme=_path_to_readme,
            **dd,
        )

    @staticmethod
    def from_file(filepath: str = None, parent_filepath: str = None):
        """Load a Model from a file.

        Args:
            filepath (str): File from which to load the model
            parent_filepath (str): Parent filename (if file is imported from another file)
        """
        fullpath = full_filepath(filepath, parent_filepath)
        raw, md_path = load_any_file(filepath, parent_filepath)
        d = raw
        if isinstance(raw, dict):
            lc_keys = lowercase_keys(raw)
            if "model" in lc_keys:
                d = raw[lc_keys["model"]]
            elif "models" in lc_keys:
                # called Model.from_file() on a model list, fallback to ModelList
                d = raw[lc_keys["models"]]
                if isinstance(d, list):
                    from modelindex.models.ModelList import ModelList
                    return ModelList(d, fullpath, md_path)

            return Model.from_dict(d, fullpath, md_path)
        else:
            raise ValueError(f"Expected a model dict, but got "
                             f"something else in file '{fullpath}'")

    def _readme_is_filepath(self):
        return self.readme and self.readme.endswith(".md") and len(self.readme) < 256

    def build_full_model(self, col):
        # Builds a full model based on the parent collection
        # col: Collection
        model_full = copy.deepcopy(col)
        self_copy = copy.deepcopy(self)

        # Merge from src to dest dictionary by this key
        def merge_by_key(d_dest, d_src):
            for key in d_src.keys():
                # if doesn't exist, just copy over
                if key not in d_dest:
                    d_dest[key] = d_src[key]
                else:
                    # if exists try to merge dicts and lists
                    if isinstance(d_dest[key], list):
                        if isinstance(d_src[key], list):
                            d_dest[key].extend(d_src[key])
                        else:
                            d_dest[key].append(d_src[key])
                    elif isinstance(d_dest[key], dict):
                        if isinstance(d_src[key], dict):
                            # copy values that don't exist
                            for k, v in d_src[key]:
                                if k not in d_dest[key]:
                                    d_dest[key][k] = v
                        else:
                            d_dest[key] = d_src[key]
                    else:
                        # overwrite if not a list or dict
                        d_dest[key] = d_src[key]

        # merge all fields from this model
        for key, value in self_copy.data.items():
            if key == "Metadata":
                if isinstance(model_full.metadata, Metadata) and isinstance(value, Metadata):
                    merge_by_key(model_full.metadata.data, value.data)
                else:
                    model_full.metadata = value
            elif key == "Results":
                if isinstance(model_full.results, ResultList) and isinstance(value, ResultList):
                    model_full.results.data.extend(value.data)
                else:
                    model_full.results = value
            else:
                # copy over if it doesn't exist in the collection
                model_full.data[key] = value

        self._full_model = model_full
        return model_full

    def readme_content(self):
        """Get the content of the README file (instead of just the path as returned by .readme())"""

        if not self.readme:
            return None
        elif self._path_to_readme:
            with open(self.filepath, "r") as f:
                return f.read()
        elif self._readme_is_filepath():
            if self.filepath:
                fullpath = full_filepath(self.readme, self.filepath)
            else:
                fullpath = self.readme
            with open(fullpath, "r") as f:
                return f.read()
        else:
            return self.readme

    # Getters
    @property
    def name(self):
        """Get the model name"""
        return self.data.get("Name", None)

    @property
    def paper(self):
        """Get the model paper"""
        return self.data.get("Paper", None)

    @property
    def code(self):
        """Get the URL to code"""
        return self.data.get("Code", None)

    @property
    def weights(self):
        """Get the URL to weights"""
        return self.data.get("Weights", None)

    @property
    def config(self):
        """Get the URL to the config file"""
        return self.data.get("Config", None)

    @property
    def readme(self):
        """Get the path to the model README"""
        return self.data.get("README", None)

    @property
    def metadata(self):
        """Get the metadata as a Metadata object"""
        return self.data.get("Metadata", None)

    @property
    def results(self):
        """Get the results as a Result object"""
        return self.data.get("Results", None)

    @property
    def in_collection(self):
        """Get the name of the collection of which this model is part of."""
        return self.data.get("In Collection", None)

    @property
    def image(self):
        """Get the path or URL to the image for the model"""
        return self.data.get("Image", None)

    @property
    def full_model(self):
        """Get the model with all the inherited properties from the collection (read-only property)"""
        return self._full_model

    # Setters
    @name.setter
    def name(self, value):
        self.data["Name"] = value

    @paper.setter
    def paper(self, value):
        self.data["Paper"] = value

    @code.setter
    def code(self, value):
        self.data["Code"] = value

    @weights.setter
    def weights(self, value):
        self.data["Weights"] = value

    @config.setter
    def config(self, value):
        self.data["Config"] = value

    @readme.setter
    def readme(self, value):
        self.data["README"] = value

    @in_collection.setter
    def in_collection(self, value):
        self.data["In Collection"] = value

    @image.setter
    def image(self, value):
        self.data["Image"] = value

    @metadata.setter
    def metadata(self, value):
        if value is not None and not isinstance(value, Metadata) and not isinstance(value, str):
            self.data["Metadata"] = Metadata.from_dict(value)
        else:
            self.data["Metadata"] = value

    @results.setter
    def results(self, value):
        if value is not None and not isinstance(value, ResultList) and not isinstance(value, str):
            self.data["Results"] = ResultList(value)
        else:
            self.data["Results"] = value



import os
from typing import Dict

from modelindex.consts import MODEL_INDEX_ROOT_FILE
from modelindex.models.Model import Model
from modelindex.models.ModelIndex import ModelIndex
from modelindex.models.ModelList import ModelList
from modelindex.models.Result import Result
from modelindex.models.Metadata import Metadata
from modelindex.models.ResultList import ResultList
from modelindex.utils import load_any_file


def has_at_least_one_models_key(lc_keys):
    """Check if there is at least one optional Models dict key in lowercase keys
    """

    model_keys = {"paper",
                  "code",
                  "weights",
                  "config",
                  "readme",
                  "metadata",
                  "results",
                  "incollection"}

    return len(model_keys.intersection(set(lc_keys))) >= 1


def load_based_on_dict_field_guess(raw: Dict, path: str, md_path: str, is_root: bool = False):
    """Load from a dict by guessing it's type based on keys in the dict.

    Args:
        raw (dict): The dictionary we want to load in
        path (str): The path it was read from
        md_path (str): Path to the Markdown file (if dict coming from there)
        is_root (bool): If this is the root for the whole project

    """
    keys_lowercase = [k.lower() for k in raw.keys()]
    if "models" in keys_lowercase \
            or "collections" in keys_lowercase \
            or "import" in keys_lowercase:
        obj = ModelIndex.from_dict(raw, path, is_root=is_root)
    elif "task" in keys_lowercase and "dataset" in keys_lowercase and "metrics" in keys_lowercase:
        obj = Result.from_dict(raw, path)
    elif "name" in keys_lowercase and has_at_least_one_models_key(keys_lowercase):
        obj = Model.from_dict(raw, path, md_path)
    else:
        obj = Metadata.from_dict(raw, path)

    return obj


def load(path: str = "model-index.yml"):
    """Load the model index.

       Args:
            path (str): Path to the file to load, or a directory containing model-index.yml.

       Returns:
            ModelIndex instance that has been loaded from the disk.
    """

    if os.path.isdir(path):
        new_path = os.path.join(path, MODEL_INDEX_ROOT_FILE)
        # if model-index.yml doesn't exist try model-index.yaml
        if not os.path.exists(new_path):
            yaml_ext = os.path.join(path, "model-index.yaml")
            if os.path.exists(yaml_ext):
                new_path = yaml_ext
        path = new_path

    raw, md_path = load_any_file(path)

    # make sure the input is a dict
    if not isinstance(raw, dict) and not isinstance(raw, list):
        raise ValueError(f"Expected the file '{path}' to contain a dict or a list, but it doesn't.")

    # Guess the type based on dict entries
    obj = None
    if isinstance(raw, dict) and raw:
        obj = load_based_on_dict_field_guess(raw, path, md_path, is_root=True)
    elif isinstance(raw, list) and raw:
        obj_guess = load_based_on_dict_field_guess(raw[0], path, md_path, is_root=True)
        if isinstance(obj_guess, Result):
            return ResultList(raw, path)
        elif isinstance(obj_guess, Model):
            return ModelList(raw, path)

    if obj is None:
        raise ValueError(f"Unrecognized format in file '{path}'")

    return obj



import os
from typing import List, Dict

import yaml
import json
import markdown
import glob
from html.parser import HTMLParser


class CommentHTMLParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.extracted_comments = []

    def handle_comment(self, data):
        self.extracted_comments.append(data)


def lowercase_keys(d: dict):
    """Map keys to lowercase, e.g. it will map a dict with a MODELS key into
    {"models": "MODELS"}. This allows for a unified way of naming custom-case keys.
    """
    out = {}
    for k in d.keys():
        out[k.lower()] = k

    return out


def load_raw_from_markdown(path: str):
    """Load raw metadata from a markdown file

        Args:
            path (str): Path to the markdown file
    """

    with open(path, "r") as f:
        md = markdown.markdown(f.read())

    parser = CommentHTMLParser()
    parser.feed(str(md))
    comments = parser.extracted_comments

    # try to interpret as a yaml file
    metadata = []
    for comment in comments:
        try:
            parsed = yaml.load(comment, Loader=yaml.SafeLoader)
            if isinstance(parsed, dict):
                lc_keys = lowercase_keys(parsed)
                append = False
                if "type" in lc_keys and (
                        parsed[lc_keys["type"]].lower() == "model-index"
                        or
                        parsed[lc_keys["type"]].lower() == "modelindex"
                ):
                    append = True
                elif "models" in lc_keys or "collections" in lc_keys or "name" in lc_keys:
                    # Guess that this is the right thing to import...
                    append = True

                if append:
                    metadata.append(parsed)
        except Exception:
            pass

    # if there are many metadata entries, merge them
    meta_dict = {}
    for m in metadata:
        for key, value in m.items():
            if key not in meta_dict:
                meta_dict[key] = value
            else:
                cur = meta_dict[key]
                if not isinstance(cur, list):
                    cur = [cur]
                if isinstance(value, list):
                    cur.extend(value)
                else:
                    cur.append(value)

                meta_dict[key] = cur

    return meta_dict


def full_filepath(path: str, cur_filepath: str = None):
    """Return the full file path based on a possibly relative path (path)
    and the current file path (cur_filepath)
    """
    if cur_filepath is not None:
        dirname = os.path.dirname(cur_filepath)
        if dirname:
            path = os.path.join(dirname, path)

    return path


def expand_wildcard_path(path: str, cur_filepath: str = None):
    if "*" not in path and "?" not in path:
        return [path]
    else:
        # we need absolute paths for glob
        path = full_filepath(path, cur_filepath)
        fullpaths = glob.glob(path)

        # but we'll return relative paths so convert back
        if cur_filepath is not None:
            base_path = os.path.dirname(cur_filepath)
            return sorted([os.path.relpath(fp, base_path) for fp in fullpaths])
        else:
            return sorted(fullpaths)


def load_any_files_wildcard(path: str, cur_filepath: str = None):
    """Load any supported file format, supporting wildcard matches as well.

    Args:
        path (str): Path to files (including any wildcards)
        cur_filepath (str): The path to the current file from which we
                            are loading the new file
                            (important for relative paths)

    """

    out = []
    # to do globs we need the absolute path
    for filepath in expand_wildcard_path(path, cur_filepath):
        out.append(load_any_file(filepath, cur_filepath))

    return out


def load_any_file(path: str, cur_filepath: str = None):
    """Load any of the supported file formats

        Args:
            path (str): Path to the file to load
            cur_filepath (str): The path to the current file from which we
                                are loading the new file
                                (important for relative paths)

        Returns:
            (raw, md_path) - where raw is the loaded content of the file (normally a dict)
            and md_path is the file to the markdown path (if loaded from markdown)
    """

    relative_path = path
    path = full_filepath(path, cur_filepath)

    if not os.path.exists(path):
        raise IOError(f"File '{path}' does not exist.")

    md_path = None
    if path.endswith(".yml") or path.endswith(".yaml"):
        with open(path, "r") as f:
            raw = yaml.load(f, Loader=yaml.SafeLoader)
    elif path.endswith(".json"):
        with open(path, "r") as f:
            raw = json.load(f)
    elif path.endswith(".md"):
        md_path = relative_path
        raw = load_raw_from_markdown(path)
    else:
        raise ValueError(f"File type at path '{path}' not recognized. We support YAML (.yml or .yaml), "
                         f"JSON (.json) and markdown (.md) files.")

    return raw, md_path


def merge_lists_data(list_objs: List):
    """Take multiple e.g. ResultList objects and merge them together into one
       by concatenating the data lists.
    """
    if list_objs:
        objs1 = list_objs[0]
        # Merge data from all files
        if len(list_objs) > 1:
            for i in range(1, len(list_objs)):
                objs1.data.extend(list_objs[i].data)
        return objs1
    else:
        return list_objs


def merge_dicts(dict_list: List[Dict]):
    if dict_list:
        if len(dict_list) == 1:
            return dict_list[0]
        else:
            # merge dictionaries
            out = {}
            for d in dict_list:
                out.update(d)
            return out
    else:
        return None



import os

from typing import List, Union, Dict, Set
from ordered_set import OrderedSet


class BaseModelIndex:

    COMMON_FIELDS = []

    def __init__(self,
                 data: Union[List, Dict] = None,
                 filepath: str = None,
                 check_errors: Union[List, Set, OrderedSet] = None,
                 ):
        if data is None:
            data = {}

        self.filepath = filepath
        self.data = data
        self.check_errors = OrderedSet(check_errors)
        self._iterator_inx = 0

        # always call checking
        self._check(silent=True)

    def _check(self, silent: bool = False):
        pass

    def collect_check_errors(self, display_type: str = None):
        clname = type(self).__name__

        if display_type is None:
            display_type = clname

        collected = [
            {
                "filepath": self.filepath,
                "errors": self.check_errors,
                "type": display_type,
            }
        ]
        if self.data:
            objs = None
            if isinstance(self.data, list):
                # If it's a list, enumerate the position
                proper_name = clname
                if clname == "ResultList":
                    proper_name = "Results"
                elif clname == "ModelList":
                    proper_name = "Models"
                elif clname == "CollectionList":
                    proper_name = "Collections"

                objs = {}
                for i in range(0, len(self.data)):
                    objs[f"{proper_name}[{i}]"] = self.data[i]

            elif isinstance(self.data, dict):
                # if dictionary, just use it
                objs = self.data

            # recursively collect all errors
            if objs:
                for name, obj in objs.items():
                    if isinstance(obj, BaseModelIndex):
                        collected.extend(obj.collect_check_errors(name))

        return collected

    def check(self, silent=False):
        """Check if the mandatory fields are present and if file references are valid.

        Args:
            silent (bool): If to return a list of errors without printing them out

        Returns:
            If silent: list of errors

        """
        errors = self.collect_check_errors()

        # only keep non-empty
        errors = [e for e in errors if e["errors"]]

        errors_formatted = []
        for e in errors:
            for msg in e["errors"]:
                fp = e["filepath"]
                if self.filepath:
                    # generate relative paths if possible
                    fp = os.path.relpath(fp, os.path.dirname(self.filepath))

                errors_formatted.append(
                    "%s: %s: %s" % (
                        fp,
                        e["type"],
                        msg
                    )
                )

        if not silent:
            for e in errors_formatted:
                print(e)
            # Return false if there are errors
            if not errors_formatted:
                print("All good!")
        else:
            return errors_formatted

    def _str_for_dict_with_padding(self, padding=2):
        """Produce a nice string for __str__ that describes this object.

        Args:
            padding (int): number of spaces to indent
        """
        # padding at this level of indentation
        pad = " "*padding
        last_pad = " "*(padding-2)

        # concatenate common fields
        out = []
        d = self.data.copy()
        for field in self.COMMON_FIELDS:
            if field in d:
                val = d.pop(field)
                if isinstance(val, BaseModelIndex):
                    val_str = val._str_with_padding(padding+2)
                else:
                    val_str = str(val)
                out.append(f"{field}={val_str}")

        # add any remaining
        if d:
            out.append(f"custom={d}")

        if self.filepath:
            out.append(f"_filepath={self.filepath}")

        name = type(self).__name__

        if out:
            return f"{name}(\n{pad}%s\n{last_pad})" % f",\n{pad}".join(out)
        else:
            return f"{name}()"

    def _str_with_padding(self, padding=2):
        # if dict use the default
        if isinstance(self.data, dict):
            return self._str_for_dict_with_padding(padding)
        elif isinstance(self.data, list):
            # if list assume it's a list of dicts
            out = []
            for d in self.data:
                if isinstance(d, BaseModelIndex):
                    out.append(d._str_for_dict_with_padding(padding+2))

            pad = " " * padding
            last_pad = " " * (padding - 2)
            if out:
                return f"[\n{pad}%s,\n{last_pad}]" % f",\n{pad}".join(out)
            else:
                return f"[]"

        else:
            return super().__str__()

    def __str__(self):
        return self._str_with_padding()

    def __repr__(self):
        return self.__str__()



from ordered_set import OrderedSet

from modelindex.models.BaseModelIndex import BaseModelIndex
from modelindex.models.CollectionList import CollectionList
from modelindex.models.ModelList import ModelList
from modelindex.utils import lowercase_keys, load_any_file, full_filepath, load_any_files_wildcard, \
    expand_wildcard_path, merge_lists_data


class ModelIndex(BaseModelIndex):
    """ModelIndex is the root object for the whole model index.
    """
    COMMON_FIELDS = [
        "Models",
        "Collections",
    ]

    def __init__(self,
                 data: dict = None,
                 filepath: str = None,
                 _path_to_readme: str = None,
                 is_root: bool = False,
                 ):
        """
        Args:
            data (dict): The root model index as a dictionary
            filepath (str): The path from which it was loaded
            _path_to_readme (str): The path to the readme file (if loaded from there)
            is_root (bool): If this is the root ModelIndex instance for the whole project
        """

        check_errors = OrderedSet()

        if data is None:
            data = {}

        d = {
            "Models": ModelList(_filepath=filepath),
            "Collections": CollectionList(_filepath=filepath),
        }
        lc_keys = lowercase_keys(data)
        if "models" in lc_keys:
            models = data[lc_keys["models"]]
            # Syntax: Models: <path to file(s)>
            if models is not None and isinstance(models, str):
                models_list = []
                for model_file in expand_wildcard_path(models, filepath):
                    try:
                        models_list.append(ModelList.from_file(model_file, filepath))
                    except (IOError, ValueError) as e:
                        check_errors.add(str(e))
                models = merge_lists_data(models_list)
            # Syntax: Models: list[ model dict ]
            elif models is not None and not isinstance(models, ModelList):
                models = ModelList(models, filepath, _path_to_readme)

            d["Models"] = models

        if "collections" in lc_keys:
            collections = data[lc_keys["collections"]]
            # Syntax: Collections: <path to file(s)>
            if collections is not None and isinstance(collections, str):
                collections_list = []
                for model_file in expand_wildcard_path(collections, filepath):
                    try:
                        collections_list.append(CollectionList.from_file(model_file, filepath))
                    except (IOError, ValueError) as e:
                        check_errors.add(str(e))
                collections = merge_lists_data(collections_list)
            # Syntax: Collections: list[ model dict ]
            elif collections is not None and not isinstance(collections, CollectionList):
                collections = CollectionList(collections, filepath, _path_to_readme)

            d["Collections"] = collections

        if "import" in lc_keys:
            imp = data[lc_keys["import"]]

            if not isinstance(imp, list):
                imp = list(imp)

            for import_file in imp:
                try:
                    for relpath in expand_wildcard_path(import_file, filepath):
                        raw, md_name = load_any_file(relpath, filepath)
                        fullpath = full_filepath(relpath, filepath)
                        mi = ModelIndex.from_dict(raw, fullpath, md_name)
                        if mi.models:
                            for model in mi.models:
                                d["Models"].add(model)

                            for col in mi.collections:
                                d["Collections"].add(col)
                except (IOError, ValueError) as e:
                    check_errors.add(str(e))

        super().__init__(
            data=d,
            filepath=filepath,
            check_errors=check_errors,
        )

        self.lc_keys = lowercase_keys(data)
        self.is_root = is_root
        if is_root:
            self.build_models_with_collections()

    def build_models_with_collections(self):
        # Apply the metadata inheritance from the collection

        col_by_name = {}
        for col in self.collections:
            col_by_name[col.name] = col

        for model in self.models:
            col_name = model.in_collection
            if col_name:
                if col_name in col_by_name:
                    col = col_by_name[col_name]
                    model.build_full_model(col)

                else:
                    model.check_errors.add(
                        f"Invalid collection name `{col_name}`"
                    )


    @staticmethod
    def from_dict(d: dict, filepath: str = None, _path_to_readme: str = None, is_root: bool = False):
        """Construct a ModelIndex from a dictionary

        Args:
            data (dict): The root model index as a dictionary
            filepath (str): The path from which it was loaded
            _path_to_readme (str): Path to the README.md file if loaded from there
            is_root (str): If this is the root ModelIndex for the whole project
        """
        return ModelIndex(d, filepath, _path_to_readme, is_root)

    @property
    def models(self) -> ModelList:
        """Get the list of models in the ModelIndex."""
        return self.data["Models"]

    @models.setter
    def models(self, value):
        """Set the list of models in the ModelIndex."""
        self.data["Models"] = value

    @property
    def collections(self) -> CollectionList:
        """Get the list of collections in the ModelIndex."""
        return self.data["Collections"]

    @collections.setter
    def collections(self, value):
        """Set the list of collections in the ModelIndex"""
        self.data["Collections"] = value




from typing import List, Union, Dict

from ordered_set import OrderedSet

from modelindex.models.BaseModelIndex import BaseModelIndex
from modelindex.models.Result import Result
from modelindex.utils import full_filepath, load_any_file, lowercase_keys


class ResultList(BaseModelIndex):
    """ResultList is a list of Result objects."""
    def __init__(self,
                 results: Union[List[Union[Dict, Result, str]], Result, Dict] = None,
                 _filepath: str = None,
                 ):
        """
        Args:
            results (list, Result, dict): Either a list of results, a single Result object or a dict representing a result
            _filepath (str): path to the file where the data is coming from
        """

        check_errors = OrderedSet()

        if results is None:
            results = []

        if isinstance(results, Result) or isinstance(results, dict):
            results = [results]

        results_parsed = []
        for r in results:
            if isinstance(r, str):
                try:
                    results_parsed.append(Result.from_file(r, _filepath))
                except (IOError, ValueError) as e:
                    check_errors.add(str(e))
            elif isinstance(r, Result):
                results_parsed.append(r)
            else:
                results_parsed.append(Result.from_dict(r, _filepath))

        super().__init__(
            data=results_parsed,
            filepath=_filepath,
            check_errors=check_errors,
        )

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __iter__(self):
        self._iterator_inx = 0
        return self

    def __next__(self):
        if self._iterator_inx < len(self.data):
            self._iterator_inx += 1
            return self.data[self._iterator_inx - 1]
        else:
            raise StopIteration

    def __len__(self):
        return len(self.data)

    @staticmethod
    def from_file(filepath: str = None, parent_filepath: str = None):
        """Load a ResultList from a file.

        Args:
            filepath (str): File from which to load the result list
            parent_filepath (str): Parent filename (if file is imported from another file)
        """
        fullpath = full_filepath(filepath, parent_filepath)
        raw, md_path = load_any_file(filepath, parent_filepath)
        d = raw

        if isinstance(d, list):
            return ResultList(d, fullpath)
        elif isinstance(d, dict):
            lc_keys = lowercase_keys(raw)
            if "results" in lc_keys:
                return ResultList(d[lc_keys["results"]], fullpath)

        raise ValueError(f"Expected a list of results, but got something else"
                         f"in file {fullpath}")



import os
import click

from modelindex.load_model_index import load


@click.group()
def cli():
    """model-index command line client."""
    pass


@cli.command()
@click.argument("filepath", default="")
def check(filepath):
    """Check if the file syntax is valid"""
    if filepath == "":
        filepath = "model-index.yml"
        if not os.path.isfile(filepath):
            click.echo("ERROR: model-index.yml not found in current directory.")
            return

    click.echo(f'Checking {filepath}')
    mi = load(filepath)
    mi.check()







import modelindex
import pytest
from modelindex import Metadata
from modelindex.models.Collection import Collection
from modelindex.models.CollectionList import CollectionList
from modelindex.models.Model import Model
from modelindex.models.ModelList import ModelList
from modelindex.models.Result import Result
from modelindex.models.ResultList import ResultList
from modelindex.models.ModelIndex import ModelIndex
import copy


def test_deepcopy():
    mi = modelindex.load("tests/test-mi/03_col")

    m1 = mi.models[0]
    m2 = copy.deepcopy(m1)

    m2.name = "New name"
    assert m1.name != m2.name
    assert m2.name == "New name"

    m2.results[0].task = "New task"
    assert m1.results[0].task != m2.results[0].task
    assert m2.results[0].task == "New task"

    m2.results.data.append(Result(task="", dataset="", metrics={}))
    assert len(m1.results) == 1
    assert len(m2.results) == 2

    m2.metadata.flops = 10
    assert m1.metadata.flops != m2.metadata.flops
    assert m2.metadata.flops == 10


def test_col_merge():
    mi = modelindex.load("tests/test-mi/17_collections_merge")

    m1 = mi.models[0].full_model
    m2 = mi.models[1].full_model

    assert m1.metadata.training_data == "ImageNet"
    assert m2.metadata.training_data == "Reddit"

    assert len(m1.metadata.training_techniques) == 4
    assert len(m2.metadata.training_techniques) == 5
    assert m2.metadata.training_techniques[-1] == "Transformers"

    assert m1.readme == "docs/inception-v3-readme.md"
    assert m2.readme == "docs/inception-v3-readme-120.md"

    mi = modelindex.load("tests/test-mi/17_collections_merge/mi2.yml")

    m1 = mi.models[0].full_model
    m2 = mi.models[1].full_model

    assert len(m1.results) == 2
    assert len(m2.results) == 2

    assert m1.results[0].metrics["Top 1 Accuracy"] == "11%"
    assert m2.results[0].metrics["Top 1 Accuracy"] == "11%"
    assert m1.results[1].metrics["Top 1 Accuracy"] == "74.67%"
    assert m2.results[1].metrics["Top 1 Accuracy"] == "75.1%"

    mi = modelindex.load("tests/test-mi/17_collections_merge/mi3.yml")

    err = mi.check(silent=True)

    assert len(err) == 2
    assert "Inception v3-1" in err[0]

    m1 = mi.models[0].full_model
    m2 = mi.models[1].full_model

    assert m1.metadata.training_data is None
    assert m2.metadata.training_data == "Reddit"


from modelindex.models.Model import Model
from modelindex.utils import full_filepath, load_any_file, lowercase_keys


class Collection(Model):
    """A collection of models with common characteristics."""

    COMMON_FIELDS = [
        "Name",
        "Metadata",
        "Results",
        "Paper",
        "Code",
        "Weights",
        "Config",
        "README",
    ]

    @staticmethod
    def from_file(filepath: str = None, parent_filepath: str = None):
        """Load a Collection from a file.

        Args:
            filepath (str): File from which to load the collection
            parent_filepath (str): Parent filename (if file is imported from another file)
        """
        fullpath = full_filepath(filepath, parent_filepath)
        raw, md_path = load_any_file(filepath, parent_filepath)
        d = raw
        if isinstance(raw, dict):
            lc_keys = lowercase_keys(raw)
            if "collection" in lc_keys:
                d = raw[lc_keys["collection"]]
            elif "collections" in lc_keys:
                # called Collection.from_file() on a collection list, fallback to CollectionList
                d = raw[lc_keys["collections"]]
                if isinstance(d, list):
                    from modelindex.models.CollectionList import CollectionList
                    return CollectionList(d, fullpath)

            return Collection.from_dict(d, fullpath, md_path)
        else:
            raise ValueError(f"Expected a collection dict, but got "
                             f"something else in file '{fullpath}'")




from typing import List, Union, Dict

from ordered_set import OrderedSet

from modelindex.models.BaseModelIndex import BaseModelIndex
from modelindex.models.Collection import Collection
from modelindex.utils import full_filepath, load_any_file, lowercase_keys, expand_wildcard_path


class CollectionList(BaseModelIndex):
    """A list of Collection objects."""
    def __init__(self,
                 collections: Union[List[Union[Collection, Dict]], Collection, Dict] = None,
                 _filepath: str = None,
                 _path_to_readme: str = None,
                 ):
        """
        Args:
            collections (list,Collection,dict): Either a list of Collection objects, a single Collection object or a
                                                dict representing a Collection
            _filepath (str): The file path to where the data was loaded from
            _path_to_readme (str): Path to README if it was loaded from there

        """

        if collections is None:
            collections = []

        if isinstance(collections, Collection) or isinstance(collections, dict):
            collections = [collections]

        models_parsed = []
        check_errors = OrderedSet()
        for m in collections:
            if isinstance(m, str):
                # link to collection file, support wildcards
                for model_file in expand_wildcard_path(m, _filepath):
                    try:
                        model = Collection.from_file(model_file, _filepath)
                        if isinstance(model, Collection):
                            models_parsed.append(model)
                        elif isinstance(model, CollectionList):
                            models_parsed.extend(model)
                    except (IOError, ValueError) as e:
                        check_errors.add(str(e))
            elif isinstance(m, Collection):
                models_parsed.append(m)
            else:
                models_parsed.append(Collection.from_dict(m, _filepath, _path_to_readme))

        super().__init__(
            data=models_parsed,
            filepath=_filepath,
            check_errors=check_errors,
        )

    def __getitem__(self, key):
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __iter__(self):
        self._iterator_inx = 0
        return self

    def __next__(self):
        if self._iterator_inx < len(self.data):
            self._iterator_inx += 1
            return self.data[self._iterator_inx - 1]
        else:
            raise StopIteration

    def __len__(self):
        return len(self.data)

    @property
    def collections(self):
        """Get the list of Collection objects"""
        return self.data

    def add(self, col: Union[Collection, Dict], _filepath: str = None):
        """Add a Collection to the list.

        Args:
            col (Collection, dict): Either a Collection or a dict representing a collection
            _filepath (str): The path from which it was loaded

        """
        col_filepath = _filepath if _filepath is not None else self.filepath
        if isinstance(col, dict):
            self.data.append(Collection.from_dict(col, col_filepath))
        elif isinstance(col, Collection):
            self.data.append(col)

    @staticmethod
    def from_file(filepath: str = None, parent_filepath: str = None):
        """Load a Collection from a file.

        Args:
            filepath (str): File from which to load the collection
            parent_filepath (str): Parent filename (if file is imported from another file)
        """

        fullpath = full_filepath(filepath, parent_filepath)
        raw, md_path = load_any_file(filepath, parent_filepath)
        d = raw

        if isinstance(d, list):
            return CollectionList(d, fullpath)
        elif isinstance(d, dict):
            lc_keys = lowercase_keys(raw)
            if "collections" in lc_keys:
                return CollectionList(d[lc_keys["collections"]], fullpath)

        raise ValueError(f"Expected a list of collections, but got something else"
                         f"in file {fullpath}")



import modelindex
import pytest
from modelindex import Metadata
from modelindex.models.Collection import Collection
from modelindex.models.CollectionList import CollectionList
from modelindex.models.Model import Model
from modelindex.models.ModelList import ModelList
from modelindex.models.Result import Result
from modelindex.models.ResultList import ResultList
from modelindex.models.ModelIndex import ModelIndex


def test_examples():
    mi = modelindex.load("examples/option1-markdown")

    assert len(mi.collections) == 2
    assert len(mi.models) == 3

    assert mi.collections[0].name == "AlexNet"
    assert mi.collections[1].name == "ResNet"

    assert mi.models[0].code == "https://github.com/pytorch/vision/blob/5a315453da5089d66de94604ea49334a66552524/torchvision/models/alexnet.py#L53"
    assert mi.models[1].code == "https://github.com/pytorch/vision/blob/5a315453da5089d66de94604ea49334a66552524/torchvision/models/resnet.py#L304"
    assert mi.models[2].code == "https://github.com/pytorch/vision/blob/5a315453da5089d66de94604ea49334a66552524/torchvision/models/resnet.py#L316"

    mi = modelindex.load("examples/option2-yaml")

    assert len(mi.collections) == 2
    assert len(mi.models) == 3

    assert mi.collections[0].name == "AlexNet"
    assert mi.collections[1].name == "ResNet"

    assert mi.models[0].metadata.flops == 1429383808
    assert mi.models[1].metadata.flops == 15667943424
    assert mi.models[2].metadata.flops == 23117674496



import modelindex
import pytest
from modelindex import Metadata
from modelindex.models.Collection import Collection
from modelindex.models.CollectionList import CollectionList
from modelindex.models.Model import Model
from modelindex.models.ModelList import ModelList
from modelindex.models.Result import Result
from modelindex.models.ResultList import ResultList
from modelindex.models.ModelIndex import ModelIndex


def test_results_check():
    mi = modelindex.load("tests/test-mi/06_import_meta/")

    s = str(mi.models[0].metadata)
    assert "Epochs=90" in s
    s = str(mi.models[1].metadata)
    assert "Epochs=120" in s
    assert "custom={'my custom parameter': 'abc'}" in s

    s = str(mi.models[0])
    assert s == """Model(
  Name=Inception v3 - 90 epochs,
  Metadata=Metadata(
    Epochs=90,
    _filepath=tests/test-mi/06_import_meta/meta/meta1.yaml
  ),
  Results=[
    Result(
      Task=Image Classification,
      Dataset=ImageNet,
      Metrics={'Top 1 Accuracy': '74.67%', 'Top 5 Accuracy': '92.1%'},
      _filepath=tests/test-mi/06_import_meta/r1.yaml
    ),
  ],
  Weights=https://download.pytorch.org/models/inception_v3_google-1a9a5a14.pth,
  In Collection=Inception v3,
  _filepath=tests/test-mi/06_import_meta/m1.yml
)"""

    s = str(mi.models[1])
    assert s == """Model(
  Name=Inception v3 - 120 epochs,
  Metadata=Metadata(
    Epochs=120,
    custom={'my custom parameter': 'abc'},
    _filepath=tests/test-mi/06_import_meta/meta/meta2.yaml
  ),
  Results=[
    Result(
      Task=Image Classification,
      Dataset=ImageNet,
      Metrics={'Top 1 Accuracy': '75.1%', 'Top 5 Accuracy': '93.1%'},
      _filepath=tests/test-mi/06_import_meta/r2.yaml
    ),
  ],
  Weights=https://download.pytorch.org/models/inception_v3_google-120-1a9a5afd.pth,
  In Collection=Inception v3,
  _filepath=tests/test-mi/06_import_meta/m2.yml
)"""


__all__ = [
    "version",
    "__version__",
    "load",
    "Metadata",
]

from modelindex.version import version, __version__

from modelindex.load_model_index import load
from modelindex.models.Metadata import Metadata




import modelindex
import pytest
from modelindex import Metadata
from modelindex.models.Collection import Collection
from modelindex.models.CollectionList import CollectionList
from modelindex.models.Model import Model
from modelindex.models.ModelList import ModelList
from modelindex.models.Result import Result
from modelindex.models.ResultList import ResultList
from modelindex.models.ModelIndex import ModelIndex


def test_results_check():
    mi = modelindex.load("tests/test-mi/16_images/")

    assert len(mi.collections) == 1
    assert len(mi.models) == 2
    err = mi.check(silent=True)

    assert len(err) == 2
    assert "Image file nonexistant.png" in err[0]

    assert mi.collections[0].image == "images/image.png"

    mi = modelindex.load("tests/test-mi/16_images/mi2.yml")

    assert len(err) == 2
    assert "Image file nonexistant.png" in err[0]

    assert mi.models[0].image == "../images/image.png"
    assert mi.models[1].image == "http://somewhere.com/external"




from click.testing import CliRunner
from modelindex.commands.cli import cli


def test_cli_invocation():
    runner = CliRunner()
    result = runner.invoke(cli)
    assert result.exit_code == 0


def test_cli_check_ok():
    runner = CliRunner()
    result = runner.invoke(cli, ["check", "tests/test-mi/11_markdown/rexnet.md"])
    assert result.exit_code == 0
    assert "Checking" in result.output
    assert "All good" in result.output


def test_cli_check_fail():
    runner = CliRunner()
    result = runner.invoke(cli, ["check", "tests/test-mi/01_base"])
    assert result.exit_code == 0
    assert "Path to README file docs/inception-v3-readme.md is not a valid file" in result.output



import modelindex
from modelindex import Metadata
from modelindex.models.Collection import Collection
from modelindex.models.CollectionList import CollectionList
from modelindex.models.Model import Model
from modelindex.models.ModelList import ModelList
from modelindex.models.Result import Result
from modelindex.models.ResultList import ResultList


def test_singlefile():
    mi = modelindex.load("tests/test-mi/01_base")

    assert "Models" in mi.data
    assert isinstance(mi.models, ModelList)


def test_metadata():

    # empty should work
    Metadata()

    # it should allow for named arguments
    meta = Metadata(
        flops=1000,
        parameters="24M"
    )

    assert meta.data == {
        "FLOPs": 1000,
        "Parameters": "24M"
    }

    # should allow for custom
    meta = Metadata(
        flops=1000,
        my_argument=123
    )

    assert meta.data == {
        "FLOPs": 1000,
        "my_argument": 123
    }

    meta.data["my_argument"] = 11
    assert meta.data == {
        "FLOPs": 1000,
        "my_argument": 11
    }

    del meta.data["my_argument"]
    assert meta.data == {
        "FLOPs": 1000,
    }

    meta = Metadata.from_dict({
        "Epochs": 120,
        "my custom parameter": "abc"
    })
    assert meta.data == {
        "Epochs": 120,
        "my custom parameter": "abc"
    }


def test_metadata_from_dict():
    m = Metadata.from_dict(
        {'FLOPs': 11462568384,
         'Parameters': 23834568,
         'Epochs': 90,
         'Batch Size': 32,
         'Training Data': 'ImageNet',
         'Training Techniques': ['RMSProp',
                                 'Weight Decay',
                                 'Gradient Clipping',
                                 'Label Smoothing'],
         'Training Resources': '8x V100 GPUs',
         'Architecture': ['Auxiliary Classifier', 'Inception-v3 Module']}
    )

    assert m.data == {'FLOPs': 11462568384,
                      'Parameters': 23834568,
                      'Epochs': 90,
                      'Batch Size': 32,
                      'Training Data': 'ImageNet',
                      'Training Techniques': ['RMSProp',
                                              'Weight Decay',
                                              'Gradient Clipping',
                                              'Label Smoothing'],
                      'Training Resources': '8x V100 GPUs',
                      'Architecture': ['Auxiliary Classifier', 'Inception-v3 Module']}

    # mixed case and _ instead of space

    m = Metadata.from_dict(
        {'flops': 11462568384,
         'Parameters': 23834568,
         'epochs': 90,
         'batch_size': 32,
         'Training Data': 'ImageNet',
         'Training Techniques': ['RMSProp',
                                 'Weight Decay',
                                 'Gradient Clipping',
                                 'Label Smoothing'],
         'Training Resources': '8x V100 GPUs',
         'Architecture': ['Auxiliary Classifier', 'Inception-v3 Module']}
    )

    assert m.data == {'FLOPs': 11462568384,
                      'Parameters': 23834568,
                      'Epochs': 90,
                      'Batch Size': 32,
                      'Training Data': 'ImageNet',
                      'Training Techniques': ['RMSProp',
                                              'Weight Decay',
                                              'Gradient Clipping',
                                              'Label Smoothing'],
                      'Training Resources': '8x V100 GPUs',
                      'Architecture': ['Auxiliary Classifier', 'Inception-v3 Module']}


def test_result_list():
    ResultList()

    r = ResultList(
        [{
            "task": "Image Classification",
            "dataset": "ImageNet",
            "metrics": {
                "Top 1 Accuracy": "88%",
                "Top 5 Accuracy": "92%"
            }
        }]
    )

    assert len(r.data) == 1
    assert isinstance(r.data[0], Result)
    assert r.data[0].data == {
            "Task": "Image Classification",
            "Dataset": "ImageNet",
            "Metrics": {
                "Top 1 Accuracy": "88%",
                "Top 5 Accuracy": "92%"
            }
        }


def test_model():
    Model(name="abc")

    m = Model.from_dict(
        {'Name': 'Inception v3',
         'Metadata': {'FLOPs': 11462568384,
                      'Parameters': 23834568,
                      'Epochs': 90,
                      'Batch Size': 32,
                      'Training Data': 'ImageNet',
                      'Training Techniques': ['RMSProp',
                                              'Weight Decay',
                                              'Gradient Clipping',
                                              'Label Smoothing'],
                      'Training Resources': '8x V100 GPUs',
                      'Architecture': ['Auxiliary Classifier', 'Inception-v3 Module']},
         'Results': [{'Task': 'Image Classification',
                      'Dataset': 'ImageNet',
                      'Metrics': {'Top 1 Accuracy': '74.67%', 'Top 5 Accuracy': '92.1%'}}],
         'Paper': 'https://arxiv.org/abs/1512.00567v3',
         'Code': 'https://github.com/rwightman/pytorch-image-models/blob/timm/models/inception_v3.py#L442',
         'Weights': 'https://download.pytorch.org/models/inception_v3_google-1a9a5a14.pth',
         "Config": "config/myconfig.json",
         'README': 'docs/inception-v3-readme.md',
         'Other': 'something'}
    )

    assert m.name == 'Inception v3'
    assert m.paper == 'https://arxiv.org/abs/1512.00567v3'
    assert m.weights == "https://download.pytorch.org/models/inception_v3_google-1a9a5a14.pth"
    assert m.readme == "docs/inception-v3-readme.md"
    assert m.code == "https://github.com/rwightman/pytorch-image-models/blob/timm/models/inception_v3.py#L442"
    assert m.config == "config/myconfig.json"
    assert isinstance(m.metadata, Metadata)
    assert m.metadata.data == {'FLOPs': 11462568384,
                      'Parameters': 23834568,
                      'Epochs': 90,
                      'Batch Size': 32,
                      'Training Data': 'ImageNet',
                      'Training Techniques': ['RMSProp',
                                              'Weight Decay',
                                              'Gradient Clipping',
                                              'Label Smoothing'],
                      'Training Resources': '8x V100 GPUs',
                      'Architecture': ['Auxiliary Classifier', 'Inception-v3 Module']}

    assert m.data["Other"] == "something"

    assert isinstance(m.results, ResultList)
    assert m.readme == 'docs/inception-v3-readme.md'

    m.name = "New name"
    assert m.name == "New name"
    assert m.data["Name"] == "New name"

    m.results = {'Task': 'Image Classification',
                 'Dataset': 'ImageNet2',
                 'Metrics': {'Top 1 Accuracy': '74.67%', 'Top 5 Accuracy': '92.1%'}}

    assert isinstance(m.results[0], Result)
    assert m.results[0].dataset == "ImageNet2"

    # setters
    m.paper = "new paper"
    assert m.paper == "new paper"

    m.code = "new code"
    assert m.code == "new code"

    m.weights = "new w"
    assert m.weights == "new w"

    m.config = "new c"
    assert m.config == "new c"

    m.readme = "readme1"
    assert m.readme == "readme1"

    m.in_collection = "in col1"
    assert m.in_collection == "in col1"

    m.metadata = {'FLOPs': 122,
                  'Parameters': 111,
                  'Epochs': 90,
                  'Batch Size': 32,
                  'Training Data': 'ImageNet',
                  'Training Techniques': ['RMSProp',
                                          'Weight Decay',
                                          'Gradient Clipping',
                                          'Label Smoothing'],
                  'Training Resources': '8x V100 GPUs',
                  'Architecture': ['Auxiliary Classifier', 'Inception-v3 Module']}

    assert isinstance(m.metadata, Metadata)
    assert m.metadata.data["FLOPs"] == 122
    assert m.metadata.flops == 122
    m.metadata.flops = 9
    assert m.metadata.flops == 9


def test_model_list():
    ModelList()

    ml = ModelList(
        {'Name': 'Inception v1',
         'Metadata': {'FLOPs': 11462568384,
                      'Parameters': 23834568,
                      'Epochs': 90,
                      'Batch Size': 32,
                      'Training Data': 'ImageNet',
                      'Training Techniques': ['RMSProp',
                                              'Weight Decay',
                                              'Gradient Clipping',
                                              'Label Smoothing'],
                      'Training Resources': '8x V100 GPUs',
                      'Architecture': ['Auxiliary Classifier', 'Inception-v3 Module']},
         'Results': [{'Task': 'Image Classification',
                      'Dataset': 'ImageNet',
                      'Metrics': {'Top 1 Accuracy': '74.67%', 'Top 5 Accuracy': '92.1%'}}],
         'Paper': 'https://arxiv.org/abs/1512.00567v3',
         'Code': 'https://github.com/rwightman/pytorch-image-models/blob/timm/models/inception_v3.py#L442',
         'Weights': 'https://download.pytorch.org/models/inception_v3_google-1a9a5a14.pth',
         'README': 'docs/inception-v3-readme.md'}
    )

    assert len(ml.models) == 1
    assert isinstance(ml.models[0], Model)
    assert ml.models[0].name == "Inception v1"


def test_collections():
    Collection(name="abc")
    CollectionList()

    cl = CollectionList(
        [{'Name': 'Inception v3',
          'Metadata': {'Training Data': 'ImageNet',
                       'Training Techniques': ['RMSProp',
                                               'Weight Decay',
                                               'Gradient Clipping',
                                               'Label Smoothing'],
                       'Training Resources': '8x V100 GPUs',
                       'Architecture': ['Auxiliary Classifier', 'Inception-v3 Module']},
          'Paper': 'https://arxiv.org/abs/1512.00567v3',
          'Code': 'https://github.com/rwightman/pytorch-image-models/blob/timm/models/inception_v3.py#L442',
          'README': 'docs/inception-v3-readme.md'}]
    )

    assert len(cl.collections) == 1
    assert cl.collections[0].name == "Inception v3"


    m = Model.from_dict(
        {'Name': 'Inception v3 - 90 epochs',
         'In Collection': 'Inception v3',
         'Metadata': {'Epochs': 90},
         'Results': [{'Task': 'Image Classification',
                      'Dataset': 'ImageNet',
                      'Metrics': {'Top 1 Accuracy': '74.67%', 'Top 5 Accuracy': '92.1%'}}],
         'Weights': 'https://download.pytorch.org/models/inception_v3_google-1a9a5a14.pth'}
    )
    assert m.in_collection == 'Inception v3'


def test_collections_load():
    mi = modelindex.load("tests/test-mi/03_col")

    assert "Models" in mi.data
    assert isinstance(mi.models, ModelList)

    assert isinstance(mi.collections, CollectionList)


def test_common_dict_list_errors():
    mi = modelindex.load("tests/test-mi/14_common")

    assert mi.models[1].metadata.epochs == 130
    assert mi.models[0].results[0].task == "Image Classification"
    assert mi.models[0].results[0].metrics["Top 1 Accuracy"] == "11.67%"



from typing import Dict

from modelindex.models.BaseModelIndex import BaseModelIndex
from modelindex.utils import lowercase_keys, load_any_file, full_filepath


class Result(BaseModelIndex):
    """Result represents a model result on a particular benchmark."""
    COMMON_FIELDS = [
        "Task",
        "Dataset",
        "Metrics",
    ]

    def __init__(self,
                 task: str,
                 dataset: str,
                 metrics: Dict,
                 _filepath: str = None,
                 ):
        """
        Args:
            task (str): The name of the ML task
            dataset (str): The name of the dataset
            metrics (dict): A dictionary of metrics
            _filepath (str): Path to the file where the data was loaded from
        """
        d = {
            "Task": task,
            "Dataset": dataset,
            "Metrics": metrics,
        }

        super().__init__(
            data=d,
            filepath=_filepath
        )

    def _check(self, silent=True):
        for field in ["Task", "Dataset", "Metrics"]:
            if field not in self.data or self.data[field] is None:
                self.check_errors.add(f"Field '{field}' is missing")

        for field in ["Task", "Dataset"]:
            if field in self.data and self.data[field] is not None and not isinstance(self.data[field], str):
                name = type(self.data[field]).__name__
                self.check_errors.add(f"The '{field}' field should be a string but got {name}")

        if "Metrics" in self.data and self.data["Metrics"]:
            m = self.data["Metrics"]
            if not isinstance(m, list) and not isinstance(m, dict):
                self.check_errors.add("The 'Metrics' fields needs to be either a dict or a list")

    @staticmethod
    def from_dict(d: Dict, _filepath: str = None):
        """Create a Result from a dict.

        Args:
            d (dict): dictionary containing result data
            _filepath (str): The file path to where the data was loaded from
        """
        lc_keys = lowercase_keys(d)

        task = None
        dataset = None
        metrics = None

        if "task" in lc_keys:
            task = d[lc_keys["task"]]
        if "dataset" in lc_keys:
            dataset = d[lc_keys["dataset"]]
        if "metrics" in lc_keys:
            metrics = d[lc_keys["metrics"]]

        return Result(
            task=task,
            dataset=dataset,
            metrics=metrics,
            _filepath=_filepath,
        )

    @staticmethod
    def from_file(filepath: str = None, parent_filepath: str = None):
        """Load a Result from a file.

        Args:
            filepath (str): File from which to load the result
            parent_filepath (str): Parent filename (if file is imported from another file)
        """
        fullpath = full_filepath(filepath, parent_filepath)
        raw, md_path = load_any_file(filepath, parent_filepath)
        d = raw
        if isinstance(raw, dict):
            lc_keys = lowercase_keys(raw)
            if "result" in lc_keys:
                d = raw[lc_keys["result"]]
            elif "results" in lc_keys:
                # called Result.from_file() on a result list, fallback to ResultList
                d = raw[lc_keys["results"]]
                if isinstance(d, list):
                    from modelindex.models.ResultList import ResultList
                    return ResultList(d, fullpath)

            return Result.from_dict(d, fullpath)
        else:
            raise ValueError(f"Expected a results dict, but got "
                             f"something else in file '{fullpath}'")

    @property
    def dataset(self):
        """Get the dataset name"""
        return self.data.get("Dataset", None)

    @property
    def task(self):
        """Get the ML task name"""
        return self.data.get("Task", None)

    @property
    def metrics(self):
        """Get the dictionary of metrics"""
        return self.data.get("Metrics", None)

    # Setters
    @dataset.setter
    def dataset(self, value):
        self.data["Dataset"] = value

    @task.setter
    def task(self, value):
        self.data["Task"] = value

    @metrics.setter
    def metrics(self, value):
        self.data["Metrics"] = value







from setuptools import setup, find_packages
from modelindex import __version__

name = "model-index"
author = "Robert Stojnic"
author_email = "hello@paperswithcode.com"
url = "https://github.com/paperswithcode/model-index"


setup(
    name=name,
    version=__version__,
    author=author,
    author_email=author_email,
    maintainer=author,
    maintainer_email=author_email,
    description="Create a source of truth for ML model results and browse it on Papers with Code",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    url=url,
    platforms=["Windows", "POSIX", "MacOSX"],
    license="MIT",
    packages=find_packages(),
    install_requires=open("requirements.txt").read().splitlines(),
    entry_points="""
        [console_scripts]
        mi=modelindex.__main__:cli
    """,
)



import modelindex
import pytest
from modelindex import Metadata
from modelindex.models.Collection import Collection
from modelindex.models.CollectionList import CollectionList
from modelindex.models.Model import Model
from modelindex.models.ModelList import ModelList
from modelindex.models.Result import Result
from modelindex.models.ResultList import ResultList
from modelindex.models.ModelIndex import ModelIndex


def test_results_check():
    mi = modelindex.load("tests/test-mi/09_check/mi1.yml")

    e = mi.models[0].results[0].check_errors
    assert len(e) == 1
    assert "dataset" in e[0].lower()

    msgs = mi.check(silent=True)
    assert len(msgs) == 1
    assert "Results[0]" in msgs[0]

    mi = modelindex.load("tests/test-mi/09_check/mi2.yml")

    e = mi.models[0].results[0].check_errors
    assert len(e) == 0

    mi = modelindex.load("tests/test-mi/09_check/mi3.yml")

    e = mi.models[0].results[0].check_errors
    assert len(e) == 2
    assert "dataset" in e[0].lower()
    assert "metrics" in e[1].lower()

    mi = modelindex.load("tests/test-mi/09_check/mi4.yml")

    e = mi.models[0].results[0].check_errors
    assert len(e) == 3
    assert "dataset" in e[0].lower()
    assert "task" in e[1].lower()
    assert "metrics" in e[2].lower()

    assert len(mi.models[0].metadata.check_errors) == 0
    assert len(mi.models[0].check_errors) == 0

    mi = modelindex.load("tests/test-mi/09_check/mi5.yml")
    assert len(mi.models[0].check_errors) == 1
    assert "what" in mi.models[0].check_errors[0]
    assert "not exist" in mi.models[0].check_errors[0]

    msgs = mi.check(silent=True)
    assert len(msgs) == 4

    mi = modelindex.load("tests/test-mi/09_check/mi6.yml")
    assert len(mi.models[0].check_errors) == 2
    assert "what" in mi.models[0].check_errors[0]
    assert "not exist" in mi.models[0].check_errors[0]
    assert "README" in mi.models[0].check_errors[1]

    assert len(mi.models[0].results.check_errors) == 1
    assert "django" in mi.models[0].results.check_errors[0]
    assert "not exist" in mi.models[0].results.check_errors[0]

    mi = modelindex.load("tests/test-mi/09_check/mi7.yml")
    assert len(mi.check_errors) == 1
    assert "not exist" in mi.check_errors[0]

    mi = modelindex.load("tests/test-mi/10_import_meta_check")

    assert len(mi.models) == 2
    assert len(mi.collections) == 1

    assert mi.models[0].filepath.endswith("meta/m1.yml")
    msgs = mi.check(silent=True)

    assert "noexist.json" in msgs[0]
    assert "wrongsubdir" in msgs[1]

    mi = modelindex.load("tests/test-mi/09_check/mi8.yml")
    assert len(mi.models[0].check_errors) == 1
    assert "what" in mi.models[0].check_errors[0]
    assert "not exist" in mi.models[0].check_errors[0]


def test_any_file():
    m1 = modelindex.load("tests/test-mi/07_import_models/model1.yaml")

    assert isinstance(m1, Model)
    assert m1.metadata.epochs == 90
    assert len(m1.results) == 1

    m2 = modelindex.load("tests/test-mi/07_import_models/model2.yaml")

    assert isinstance(m2, Model)
    assert m2.metadata.epochs == 120
    assert len(m2.results) == 1

    mi = modelindex.load("tests/test-mi/11_markdown/subdir/rexnet3.md")

    assert isinstance(mi, ModelIndex)
    assert mi.models[0].name == "RexNet3"

    rl = modelindex.load("tests/test-mi/09_check/results_list.yaml")

    assert isinstance(rl, ResultList)
    assert len(rl) == 3
    assert rl[1].metrics["Top 1 Accuracy"] == "70.67%"
    assert len(rl[2].check_errors) == 1
    assert "Metrics" in rl[2].check_errors[0]

    with pytest.raises(ValueError):
        modelindex.load("tests/test-mi/09_check/invalid_file.yml")


from typing import Union, Dict, List

from modelindex.models.BaseModelIndex import BaseModelIndex
from modelindex.utils import lowercase_keys, full_filepath, load_any_file, merge_dicts


class Metadata(BaseModelIndex):
    """Metadata for a model."""

    COMMON_FIELDS = [
        "FLOPs",
        "Parameters",
        "Epochs",
        "Batch Size",
        "Training Data",
        "Training Techniques",
        "Training Resources",
        "Architecture",
    ]

    def __init__(self,
                 flops: Union[str, int] = None,
                 parameters: Union[str, int] = None,
                 epochs: Union[str, int] = None,
                 batch_size: Union[str, int] = None,
                 training_data: [str, List] = None,
                 training_techniques: [str, List] = None,
                 training_resources: str = None,
                 architecture: [str, List] = None,
                 _filepath: str = None,
                 **kwargs,
                 ):
        """
        Args:
            flops (str,int): number of FLOPs
            parameters (str,int): total number of parameters for the model
            epochs (str,int): how many epochs the model was trained
            batch_size (str,int): batch size for the model
            training_data (str,list): one or a list of datasets used in training
            training_techniques (str,list): one or a list of training techniques
            training_resources (str): hardware used to train
            architecture (str, List): one or a list of architectures used in the model
            _filepath (str): path to the file where the data is coming from
            **kwargs: any other custom metadata
        """
        d = {
            "FLOPs": flops,
            "Parameters": parameters,
            "Epochs": epochs,
            "Batch Size": batch_size,
            "Training Data": training_data,
            "Training Techniques": training_techniques,
            "Training Resources": training_resources,
            "Architecture": architecture,
            **kwargs,
        }

        # only save non-None values
        data = {k: v for k, v in d.items() if v is not None}

        super().__init__(
            data=data,
            filepath=_filepath
        )

    def _check(self, silent=True):
        if not isinstance(self.data, dict) and not isinstance(self.data, list):
            self.check_errors.add("Metadata should be either a list or a dict")

    @staticmethod
    def from_dict(d: Dict, _filepath: str = None):
        """Construct Metadata from a dict.

        Args:
            d (dict): A dictionary of values
            _filepath (str): path to the file where the data is coming from

        """
        # be flexible to a common error where
        if isinstance(d, list) and len(d) > 0:
            d = merge_dicts(d)

        lc_keys = lowercase_keys(d)

        # common error: have metadata twice
        if len(lc_keys) == 1 and "metadata" in lc_keys:
            d = d[lc_keys["metadata"]]
            lc_keys = lowercase_keys(d)

        dd = d.copy()
        for field_name in Metadata.COMMON_FIELDS:
            key = field_name.lower()
            if key in lc_keys:
                dd[field_name] = dd.pop(lc_keys[key])

            # try with _ instead of space in the field name
            if " " in field_name:
                key = field_name.lower().replace(" ", "_")
                if key in lc_keys:
                    dd[field_name] = dd.pop(lc_keys[key])

        return Metadata(
            _filepath=_filepath,
            **dd,
        )

    @staticmethod
    def from_file(filepath: str = None, parent_filepath: str = None):
        """Load a Metadata from a file.

        Args:
            filepath (str): File from which to load the metadata
            parent_filepath (str): Parent filename (if file is imported from another file)
        """
        fullpath = full_filepath(filepath, parent_filepath)
        raw, md_path = load_any_file(filepath, parent_filepath)

        d = raw
        if isinstance(d, dict):
            return Metadata.from_dict(d, fullpath)

        raise ValueError(f"Expected a dictionary with metadata, "
                         f"but got something else in file at"
                         f"'{fullpath}'")

    # Getters
    @property
    def flops(self):
        """Get the FLOPs"""
        return self.data.get("FLOPs", None)

    @property
    def parameters(self):
        """Get number of parameters"""
        return self.data.get("Parameters", None)

    @property
    def epochs(self):
        """Get epochs"""
        return self.data.get("Epochs", None)

    @property
    def batch_size(self):
        """Get batch size"""
        return self.data.get("Batch Size", None)

    @property
    def training_data(self):
        """Get training data used"""
        return self.data.get("Training Data", None)

    @property
    def training_techniques(self):
        """Get techniques used"""
        return self.data.get("Training Techniques", None)

    @property
    def training_resources(self):
        """Get training resources used"""
        return self.data.get("Training Resources", None)

    @property
    def architecture(self):
        """Get the architecture(s) used."""
        return self.data.get("Architecture", None)

    # Setters
    @flops.setter
    def flops(self, value):
        self.data["FLOPs"] = value

    @parameters.setter
    def parameters(self, value):
        self.data["Parameters"] = value

    @epochs.setter
    def epochs(self, value):
        self.data["Epochs"] = value

    @batch_size.setter
    def batch_size(self, value):
        self.data["Batch Size"] = value

    @training_data.setter
    def training_data(self, value):
        self.data["Training Data"] = value

    @training_techniques.setter
    def training_techniques(self, value):
        self.data["Training Techniques"] = value

    @training_resources.setter
    def training_resources(self, value):
        self.data["Training Resources"] = value

    @architecture.setter
    def architecture(self, value):
        self.data["Architecture"] = value












import modelindex
from modelindex import Metadata
from modelindex.models.Collection import Collection
from modelindex.models.CollectionList import CollectionList
from modelindex.models.Model import Model
from modelindex.models.ModelList import ModelList
from modelindex.models.Result import Result
from modelindex.models.ResultList import ResultList
from modelindex.utils import expand_wildcard_path


def test_imports():
    mi = modelindex.load("tests/test-mi/04_imports")

    assert "Models" in mi.data
    assert isinstance(mi.models, ModelList)
    assert len(mi.models) == 2


def test_result_imports():
    mi = modelindex.load("tests/test-mi/05_field_imports")

    assert "Models" in mi.data
    assert isinstance(mi.models, ModelList)
    assert len(mi.models) == 2

    assert mi.models[0].results[0].data == {
        'Task': 'Image Classification',
        'Dataset': 'ImageNet',
        'Metrics': {'Top 1 Accuracy': '74.67%', 'Top 5 Accuracy': '92.1%'}}

    assert mi.models[1].results[0].data == {
        'Task': 'Image Classification',
        'Dataset': 'ImageNet',
        'Metrics': {'Top 1 Accuracy': '75.1%', 'Top 5 Accuracy': '93.1%'}}


def test_metadata_imports():
    mi = modelindex.load("tests/test-mi/06_import_meta")

    assert "Models" in mi.data
    assert isinstance(mi.models, ModelList)
    assert len(mi.models) == 2

    assert mi.models[0].metadata.data == {
        "Epochs": 90
    }

    assert mi.models[1].metadata.data == {
        "Epochs": 120,
        "my custom parameter": "abc"
    }


def test_models_imports():
    mi = modelindex.load("tests/test-mi/07_import_models")

    assert "Models" in mi.data
    assert isinstance(mi.models, ModelList)
    assert len(mi.models) == 2
    assert len(mi.collections) == 1

    assert mi.models[0].metadata.data == {
        "Epochs": 90
    }

    assert mi.models[1].metadata.data == {
        "Epochs": 120,
    }


def test_models_imports_json():
    mi = modelindex.load("tests/test-mi/08_import_models_json")

    assert "Models" in mi.data
    assert isinstance(mi.models, ModelList)
    assert len(mi.models) == 2

    assert mi.models[0].metadata.data == {
        "Epochs": 90
    }

    assert mi.models[1].metadata.data == {
        "Epochs": 111,
    }


def test_expand_wildcard_path():
    supported = expand_wildcard_path("tests/test-mi/12_wildcard_imports/*.yml")
    assert supported == ["tests/test-mi/12_wildcard_imports/model-index.yml"]

    supported = expand_wildcard_path("tests/test-mi/12_wildcard_imports/model-index.yml")
    assert supported == ["tests/test-mi/12_wildcard_imports/model-index.yml"]

    supported = expand_wildcard_path("tests/test-mi/12_wildcard_imports/abbccabbbcc")
    assert supported == ["tests/test-mi/12_wildcard_imports/abbccabbbcc"]

    supported = expand_wildcard_path("tests/test-mi/12_wildcard_imports")
    assert supported == ["tests/test-mi/12_wildcard_imports"]

    supported = expand_wildcard_path("tests/test-mi/12_wildcard_imports/models/*.yml")
    assert supported == ["tests/test-mi/12_wildcard_imports/models/m1.yml",
                         "tests/test-mi/12_wildcard_imports/models/m2.yml"]


def test_wildcard_import():
    mi = modelindex.load("tests/test-mi/12_wildcard_imports")

    assert len(mi.models) == 2


def test_wildcard_model_import():
    mi = modelindex.load("tests/test-mi/13_wildcard_model_imports")
    assert len(mi.models) == 2
    assert len(mi.collections) == 1

    mi = modelindex.load("tests/test-mi/13_wildcard_model_imports/mi2.yml")
    assert len(mi.models) == 2
    assert len(mi.collections) == 0

    mi = modelindex.load("tests/test-mi/13_wildcard_model_imports/mi3.yml")
    assert len(mi.models) == 2
    assert len(mi.collections) == 1

    mi = modelindex.load("tests/test-mi/13_wildcard_model_imports/mi4.yml")
    assert len(mi.models) == 2
    assert len(mi.collections) == 1

    mi = modelindex.load("tests/test-mi/13_wildcard_model_imports/mi5.yml")
    assert len(mi.models) == 0
    assert len(mi.collections) == 2

    mi = modelindex.load("tests/test-mi/13_wildcard_model_imports/mi6.yml")
    assert len(mi.models) == 2
    assert len(mi.collections) == 1

    mi = modelindex.load("tests/test-mi/13_wildcard_model_imports/mi7.yml")
    assert len(mi.models) == 2
    assert len(mi.collections) == 1
    assert mi.models[0].name == "Inception v3 - 90 epochs"
    assert len(mi.models[0].results) == 2
    assert mi.models[0].results[0].metrics["Top 1 Accuracy"] == "11%"
    assert mi.models[0].results[1].metrics["Top 1 Accuracy"] == "21%"


def test_models_import_wildcard():
    mi = modelindex.load("tests/test-mi/13_wildcard_model_imports/mi8.yml")

    err = mi.check(silent=True)
    assert len(err) == 1
    assert "docs/inception-v3-readme.md" in err[0]

    assert len(mi.collections) == 1
    assert len(mi.models) == 2

    assert mi.models[0].readme == "models_md/m1.md"
    assert mi.models[1].readme == "models_md/m2.md"

    mi = modelindex.load("tests/test-mi/13_wildcard_model_imports/mi9.yml")

    assert len(mi.collections) == 1
    assert len(mi.models) == 2

    assert mi.models[0].readme == "models_md/m1.md"
    assert mi.models[1].readme == "models_md/m2.md"


def test_readme_content():
    mi = modelindex.load("tests/test-mi/15_readme_contents/")

    assert len(mi.models) == 2
    assert mi.models[0].readme_content() == "This is some custom readme here."
    assert mi.models[1].readme_content() == "Another custom readme"

    mi = modelindex.load("tests/test-mi/15_readme_contents/mi2.yml")

    err = mi.check(silent=True)
    assert len(err) == 1
    assert "docs/inception-v3-readme.md" in err[0]

    assert len(mi.collections) == 1
    assert len(mi.models) == 2

    assert "something here." in mi.models[0].readme_content()
    assert "# Second here" in mi.models[1].readme_content()

